@@ -46,6 +46,7 @@ class TomlConfig:
 class GIDocConfig:
     """Load and represent the configuration for gidocgen"""
     def __init__(self, config_file=None):
+        self._favicons = []
         self._config_file = config_file
 
         self._config = {}
@@ -87,6 +88,12 @@ class GIDocConfig:
     def urlmap_file(self):
         return self.extra.get('urlmap_file')
 
+    @property
+    def urlmap_basename(self):
+        if self.urlmap_file is not None:
+            return os.path.basename(self.urlmap_file)
+        return None
+
     @property
     def version(self):
         return self.library.get('version', 'Unknown')
@@ -264,6 +271,17 @@ class GIDocConfig:
     def generator(self):
         return f"{core.version}"
 
+    @property
+    def favicons(self):
+        if not self._favicons:
+            self._favicons = [
+                os.path.basename(p)
+                for p in self.content_images
+                if os.path.basename(p).startswith('favicon')
+            ]
+
+        return self._favicons
+
 
 class GITemplateConfig:
     """Load and represent the template configuration"""
